@@ -5,8 +5,12 @@ import hashlib
 from enum import Enum
 from xmlrpc.client import Boolean
 
-from airtouch3.constants import AT3Const
-from airtouch3.helper import AT3Helper
+import airtouch3.constants as const
+from airtouch3.helper import (
+    calculate_checksum, 
+    bit8_in_byte_on, 
+    bit7_in_byte_on
+)
 
 class AT3AcMode(Enum):
     AUTO = 0
@@ -84,21 +88,22 @@ class AT3AcUnit:
     brand = -1
     temperature = -1
     temperature_sp = -1
-    _at3class_toggle_func = None
-    _at3class_toggle_sp_func = None
 
-    def __init__(self, name, number, toggle_func, toggle_sp_func):
+    def __init__(self, name, number, at3object):
         self.name = name
         self.number = number
-        self._at3class_toggle_func = toggle_func
-        self._at3class_toggle_sp_func = toggle_sp_func
+        self._at3 = at3object
 
     def toggle(self) -> bool: 
-        return self._at3class_toggle_func(self.number)
+        return self._at3.toggle_ac_unit(self.number)
     def temperature_inc(self) -> bool: 
-        return self._at3class_toggle_sp_func(self.number, AT3Command.INCREMENT)
+        return self._at3.toggle_temperature_ac_unit(self.number, AT3Command.INCREMENT)
     def temperature_dec(self) -> bool: 
-        return self._at3class_toggle_sp_func(self.number, AT3Command.DECREMENT)
+        return self._at3.toggle_temperature_ac_unit(self.number, AT3Command.DECREMENT)
+    def set_fan_speed(self, speed:AT3AcFanSpeed) -> AT3AcFanSpeed: 
+        return self._at3.set_fan_speed_ac_unit(self.number, speed)
+    def set_mode(self, mode:AT3AcMode) -> AT3AcMode:
+        return self._at3.set_mode_ac_unit(self.number, mode)
 
 class AT3Group:
     name = "Unknown"
@@ -108,21 +113,18 @@ class AT3Group:
     open_percent = -1
     temperature = -1
     temperature_sp = -1
-    _at3class_toggle_func = None
-    _at3class_toggle_pos_func = None
 
-    def __init__(self, name, number, toggle_func, toggle_pos_func):
+    def __init__(self, name, number, at3object):
         self.name = name
         self.number = number
-        self._at3class_toggle_func = toggle_func
-        self._at3class_toggle_pos_func = toggle_pos_func
-    
+        self._at3 = at3object
+
     def toggle(self) -> bool: 
-        return self._at3class_toggle_func(self.number)
+        return self._at3.toggle_group(self.number)
     def position_inc(self) -> int: 
-        return self._at3class_toggle_pos_func(self.number, AT3Command.INCREMENT)
+        return self._at3.toggle_position_group(self.number, AT3Command.INCREMENT)
     def position_dec(self) -> int: 
-        return self._at3class_toggle_pos_func(self.number, AT3Command.DECREMENT)
+        return self._at3.toggle_position_group(self.number, AT3Command.DECREMENT)
 
 class AT3TempSensor:
     name = "Unknown"
@@ -155,7 +157,7 @@ class AirTouch3:
 
         # Send a command to the Air Touch 3
         # Byte1 = 1 means request status only
-        data = self._send_recieve(AT3Const.CMD_1_STATUS, 0, 0, 0)
+        data = self._send_recieve(const.CMD_1_STATUS, 0, 0, 0)
         
         # Process the response, returning valid processing of reponse
         return self._process_response(data)
@@ -164,8 +166,8 @@ class AirTouch3:
 
         # Send command to the Air Touch 3 if valid ac unit number
         if acUnit >= 0 and acUnit < len(self.ac_units):
-            data = self._send_recieve(AT3Const.CMD_1_AC_CTRL, acUnit,
-                                        AT3Const.CMD_4_TOGGLE, 0)
+            data = self._send_recieve(const.CMD_1_AC_CTRL, acUnit,
+                                        const.CMD_4_TOGGLE, 0)
 
             # Process the response, if fails, return none to indicate error
             if not self._process_response(data): return None
@@ -178,29 +180,58 @@ class AirTouch3:
 
     def toggle_temperature_ac_unit(self, acUnit, direction:AT3Command) -> bool:
 
-        # Send command to the Air Touch 3 if valid ac unit number
-        if acUnit >= 0 and acUnit < len(self.ac_units):
+        # Invalid Ac Unit was given
+        if acUnit < 0 or acUnit >= len(self.ac_units):
+            return None
 
-            cmd = AT3Const.CMD_4_AC_TEMP_DEC
-            if direction == AT3Command.INCREMENT: 
-                cmd = AT3Const.CMD_4_AC_TEMP_INC
-            data = self._send_recieve(AT3Const.CMD_1_AC_CTRL, acUnit, cmd, 0)
+        cmd = const.CMD_4_AC_TEMP_DEC
+        if direction == AT3Command.INCREMENT: 
+            cmd = const.CMD_4_AC_TEMP_INC
+        data = self._send_recieve(const.CMD_1_AC_CTRL, acUnit, cmd, 0)
 
-            # Process the response, if fails, return none to indicate error
-            if not self._process_response(data): return None
+        # Process the response, if fails, return none to indicate error
+        if not self._process_response(data): return None
 
-            # return status of AC Unit
-            return self.ac_units[acUnit].temperature_sp
+        # return status of AC Unit
+        return self.ac_units[acUnit].temperature_sp  
+
+    def set_fan_speed_ac_unit(self, acUnit, speed:AT3AcFanSpeed) -> AT3AcFanSpeed:
 
         # Invalid Ac Unit was given
-        return None
+        if acUnit < 0 or acUnit >= len(self.ac_units):
+            return None
+
+        data = self._send_recieve(const.CMD_1_AC_CTRL, acUnit, 
+                                    const.CMD_4_AC_FAN_SPD, speed.value)
+
+        # Process the response, if fails, return none to indicate error
+        if not self._process_response(data): return None
+
+        # return status of AC Unit
+        return self.ac_units[acUnit].fan_speed
+
+    def set_mode_ac_unit(self, acUnit, mode: AT3AcMode) -> AT3AcMode:
+
+         # Invalid Ac Unit was given
+        if acUnit < 0 or acUnit >= len(self.ac_units):
+            return None
+
+        data = self._send_recieve(const.CMD_1_AC_CTRL, acUnit, 
+                                    const.CMD_4_AC_MODE, mode.value)
+
+        # Process the response, if fails, return none to indicate error
+        if not self._process_response(data): return None
+
+        # return status of AC Unit
+        return self.ac_units[acUnit].mode
+
 
     def toggle_group(self, group) -> bool:
 
         # Send command to the Air Touch 3 if valid group number
         if group >= 0 and group < len(self.groups):
-            data = self._send_recieve(AT3Const.CMD_1_GRP_CTRL, group, 
-                                        AT3Const.CMD_4_TOGGLE, 0)
+            data = self._send_recieve(const.CMD_1_GRP_CTRL, group, 
+                                        const.CMD_4_TOGGLE, 0)
 
             # Process the response, if fails, return none to indicate error
             if not self._process_response(data): return None
@@ -216,12 +247,12 @@ class AirTouch3:
         # Send command to the Air Touch 3 if valid group number
         if group >= 0 and group < len(self.groups):
 
-            cmd = AT3Const.CMD_4_GRP_POSDEC 
+            cmd = const.CMD_4_GRP_POSDEC 
             if direction == AT3Command.INCREMENT:
-                cmd = AT3Const.CMD_4_GRP_POSINC
+                cmd = const.CMD_4_GRP_POSINC
                 
-            data = self._send_recieve(AT3Const.CMD_1_GRP_CTRL, group, cmd, 
-                                        AT3Const.CMD_5_GRP_POS)
+            data = self._send_recieve(const.CMD_1_GRP_CTRL, group, cmd, 
+                                        const.CMD_5_GRP_POS)
 
             # Process the response, if fails, return none to indicate error
             if not self._process_response(data): return None
@@ -254,7 +285,7 @@ class AirTouch3:
 
         # No data received, must be a connection error, nothing to do
         # also make sure we recieved a response of length 492 bytes
-        if not response or len(response) != AT3Const.RESPONSE_LEN:
+        if not response or len(response) != const.RESPONSE_LEN:
             self.comms_status = AT3CommsStatus.ERROR
             self.comms_status = "Invalid Response Received"
             return False
@@ -263,28 +294,27 @@ class AirTouch3:
         # these are the dampers themselves, which are "grouped" 
         # into what is usually known as zones
         zones = []
-        for z in range(AT3Const.ZONES_LEN):
+        for z in range(const.ZONES_LEN):
             # MSB is zone on/off (x) and LS three bits are zone 
             # number 0-7 (y) x000_0yyy. Note the zone number for 
             # zones 1-8 is 0-7 and repeats for zones 9-16, ie 0-7
-            byte_value = response[AT3Const.DAOF_ZONE_STATE+z]
-            zones.append(AT3Helper.bit8_in_byte_on(byte_value))
+            byte_value = response[const.DAOF_ZONE_STATE+z]
+            zones.append(bit8_in_byte_on(byte_value))
 
         # Loop through all the groups, only load the number 
         # configured in the system
-        num_groups = int(response[AT3Const.DAOF_GRP_COUNT])
-        for z in range(min(AT3Const.GROUPS_LEN, num_groups)):
+        num_groups = int(response[const.DAOF_GRP_COUNT])
+        for z in range(min(const.GROUPS_LEN, num_groups)):
 
             # Group names are all fixed character length
-            stt = AT3Const.DAOF_GRP_NAME + (z * AT3Const.GRP_NAME_LEN)
-            end = stt + AT3Const.GRP_NAME_LEN
+            stt = const.DAOF_GRP_NAME + (z * const.GRP_NAME_LEN)
+            end = stt + const.GRP_NAME_LEN
             name = response[stt:end].decode().strip().strip('\x00')
 
             # Groups are stored using their number as the index
             # Try and get the group via its number, if non found, add it
             if not self.groups.get(z):
-                self.groups[z] = AT3Group(name, z, self.toggle_group, 
-                                            self.toggle_position_group)
+                self.groups[z] = AT3Group(name, z, self)
 
             # Get the group from he dict and set the name
             group = self.groups.get(z)
@@ -293,24 +323,24 @@ class AirTouch3:
             # Get the first zone in the group and base the status 
             # of the group on that zone. All zones in a group will have
             # the same status, so only need to read the first
-            byte_value = response[AT3Const.DAOF_GRP_FIRSTZONE+z]
+            byte_value = response[const.DAOF_GRP_FIRSTZONE+z]
             first_zone = int((byte_value & 0b1111_0000) >> 4)
             group.is_on = zones[first_zone]    
 
             # Mode of zone is 7th bit, if on then temp control, 
             # else percent position
-            byte_value = response[AT3Const.DAOF_GRP_PERCENT+z]
+            byte_value = response[const.DAOF_GRP_PERCENT+z]
             group.mode = AT3GroupMode.PERECENT
-            if AT3Helper.bit8_in_byte_on(byte_value):
+            if bit8_in_byte_on(byte_value):
                 group.mode = AT3GroupMode.TEMPERATURE
 
             # Temperature setpoint is bottom 5 bits 
             # (less one from setpoint, kinda wierd?!)
-            byte_value = response[AT3Const.DAOF_GRP_SETPOINT+z]
+            byte_value = response[const.DAOF_GRP_SETPOINT+z]
             group.temperature_sp = int(byte_value & 0b0001_1111) + 1
 
             # Group percent is bottom 7 bits and 1 count per 5%
-            byte_value = response[AT3Const.DAOF_GRP_PERCENT+z]
+            byte_value = response[const.DAOF_GRP_PERCENT+z]
             percent = 5*int(byte_value & 0b0111_1111)
             group.open_percent = percent if group.is_on else 0
 
@@ -318,37 +348,36 @@ class AirTouch3:
         # TODO At the moment, get data for two air cons, need to work  out 
         # how many air cons are being used and what groups are allocated 
         # to each air con
-        for a in range(AT3Const.AC_UNIT_LEN):
+        for a in range(const.AC_UNIT_LEN):
 
             # Names are straight after each other in the config file
-            stt = AT3Const.DAOF_AC1_NAME + AT3Const.AC_NAME_LEN*a
-            end = stt + AT3Const.AC_NAME_LEN
+            stt = const.DAOF_AC1_NAME + const.AC_NAME_LEN*a
+            end = stt + const.AC_NAME_LEN
             name = response[stt:end].decode().strip().strip('\x00')
 
             # AC Units are stored using their number as the index
             # Try and get the group via its number, if non found, add it
             if not self.ac_units.get(a):
-                self.ac_units[a] = AT3AcUnit(name, a, self.toggle_ac_unit, 
-                                                self.toggle_temperature_ac_unit)
+                self.ac_units[a] = AT3AcUnit(name, a, self)
 
             # Get the ac unit and set the name
             acUnit = self.ac_units.get(a)
             acUnit.name = name
             
             # Status contains on/off and error bits
-            status = response[AT3Const.DAOF_AC1_STATUS+a]
-            acUnit.is_on = AT3Helper.bit8_in_byte_on(status)
-            acUnit.has_error = AT3Helper.bit7_in_byte_on(status)
+            status = response[const.DAOF_AC1_STATUS+a]
+            acUnit.is_on = bit8_in_byte_on(status)
+            acUnit.has_error = bit7_in_byte_on(status)
 
             # Get the brand id, we need this to issue commands to the AcUnit, 
             # otherwise, who cares right?
-            acUnit.brand = int(response[AT3Const.DAOF_AC1_BRAND+a])
+            acUnit.brand = int(response[const.DAOF_AC1_BRAND+a])
 
             # Mode at in heat/cool etc
-            acUnit.mode = AT3AcMode(int(response[AT3Const.DAOF_AC1_MODE+a]))
+            acUnit.mode = AT3AcMode(int(response[const.DAOF_AC1_MODE+a]))
             
             # Fan Speed is only bottom 4 bits
-            byte_value = response[AT3Const.DAOF_AC1_FAN+a]
+            byte_value = response[const.DAOF_AC1_FAN+a]
             acUnit.fan_speed = AT3AcFanSpeed(int(byte_value & 0b0000_1111))
 
             # Get the temp control mode, but dont do anything with it
@@ -360,20 +389,20 @@ class AirTouch3:
             # on setup, a zone temp might be being used as the AC temp 
             # feedback TODO, need to reassign the temperature based on the 
             # therm mode above
-            acUnit.temperature = int(response[AT3Const.DAOF_AC1_TEMP_PV+a])
+            acUnit.temperature = int(response[const.DAOF_AC1_TEMP_PV+a])
 
             # Temperature setpoint; ignore top two bits as usual, for 
             # temp values
-            byte_value = response[AT3Const.DAOF_AC1_TEMP_SP+a]
+            byte_value = response[const.DAOF_AC1_TEMP_SP+a]
             acUnit.temperature_sp = int(byte_value & 0b0011_1111)
 
         # Touch pad information, add to sensor list
         # The Group the touch pad temperature is assigned to
-        group_id = int(response[AT3Const.DAOF_TP_GRP_ID])
+        group_id = int(response[const.DAOF_TP_GRP_ID])
         name = "Touch Pad 1"
 
         # Save sensor into the sensor list, returns the sensor object
-        byte_value = response[AT3Const.DAOF_TP_TEMP]
+        byte_value = response[const.DAOF_TP_TEMP]
         sensor = self._update_or_add_sensor(name, byte_value)
 
         # If valid group and sensor available (should always be available 
@@ -383,16 +412,16 @@ class AirTouch3:
             self.groups[group_id - 1].temperature = sensor.temperature  
 
         # Get the sensors in the system
-        for s in range(AT3Const.TEMP_SENSOR_LEN):
-            byte_value = response[AT3Const.DAOF_TEMP_SENSORS+s]
+        for s in range(const.TEMP_SENSOR_LEN):
+            byte_value = response[const.DAOF_TEMP_SENSORS+s]
             self._update_or_add_sensor(f"Sensor {s+1}", byte_value)
 
         # Load the system name and id
-        stt = AT3Const.DAOF_SYS_NAME
-        end = stt + AT3Const.SYS_NAME_LEN
+        stt = const.DAOF_SYS_NAME
+        end = stt + const.SYS_NAME_LEN
         self.name = response[stt:end].decode().strip().strip('\x00')
-        stt = AT3Const.DAOF_SYS_ID
-        end = stt + AT3Const.SYS_ID_LEN
+        stt = const.DAOF_SYS_ID
+        end = stt + const.SYS_ID_LEN
         self.id = response[stt:end].decode().strip().strip('\x00')
 
         # Successfully processed response
@@ -400,9 +429,9 @@ class AirTouch3:
 
     def _update_or_add_sensor(self, name, byte_value):
 
-        temperature = byte_value & 0b0011_1111               # Bits 0 to 6
-        available = AT3Helper.bit8_in_byte_on(byte_value)    # Bit 8
-        low_battery = AT3Helper.bit7_in_byte_on(byte_value)  # Bit 7
+        temperature = byte_value & 0b0011_1111     # Bits 0 to 6
+        available = bit8_in_byte_on(byte_value)    # Bit 8
+        low_battery = bit7_in_byte_on(byte_value)  # Bit 7
 
         # Nothing to do if not available
         if not available:
@@ -421,15 +450,16 @@ class AirTouch3:
 
     def _send_recieve(self, byte1, byte3, byte4, byte5) -> bytes:
 
-        # Should always be much less than this (ref AT3Const.RESPONSE_LEN)
+        # Should always be much less than this (ref const.RESPONSE_LEN)
         BUFFER_SIZE = 1024
 
         # Command is 12 bytes, add checksum as 13th
-        rList = [AT3Const.CMD_0, byte1, AT3Const.CMD_2, 
+        rList = [const.CMD_0, byte1, const.CMD_2, 
                     byte3, byte4, byte5, 0, 0, 0, 0, 0, 0]
-        rChk = AT3Helper.calculate_checksum(rList)
+        rChk = calculate_checksum(rList)
         rList.extend(rChk)
         arr = bytes(rList)
+        print(*rList, sep = ", ")
 
         try:
             s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
